; IDL-DODS interface.
; Written by Daniel J. Carr - Research Systems, Inc.
; Copyright 1999, UCAR
;
; Function to get a DDS variable and return it as an IDL variable.
; This funcion is called recursively for DDS Structure and Sequence
; variable types.
;
;    var_p - (Input )
;       Pointer to DDS variable.
;    var - (Output)
;       Returns the DDS variable's data.
;    name - (Output)
;       Returns the DDS variable's name.
;
;    returns -
;       Status (1=ok, 0=fail).
;
FUNCTION DODS_VAR, conn, dds, var_p, var, name

   DODS_VAR_INFO, var_p, type, var_name, pix

   IF (N_ELEMENTS(var_name) GT 0L) THEN name = var_name

   CASE type OF
      !dods_str_c: BEGIN
         status = GET_DODS_STR(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = STRTRIM(REFORM(TEMPORARY(val)), 2)
      END
      !dods_byte_c: BEGIN
         status = GET_DODS_BYTE(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_int16_c: BEGIN
         status = GET_DODS_INT16(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_uint16_c: BEGIN
         status = GET_DODS_UINT16(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_int32_c: BEGIN
         status = GET_DODS_INT32(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_uint32_c: BEGIN
         status = GET_DODS_UINT32(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_float32_c: BEGIN
         status = GET_DODS_FLOAT32(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_float64_c: BEGIN
         status = GET_DODS_FLOAT64(conn, dds, var_p, val)
         IF ((status EQ 0UL) OR (N_ELEMENTS(val) LE 0L)) THEN RETURN, 0UL
         var = REFORM(TEMPORARY(val))
      END
      !dods_array_c: BEGIN
         d_names = BYTARR(65, 8)
         sss = LONARR(3, 8)
         status = GET_DODS_ARRAY(conn, dds, var_p, arr, d_names, sss)
         IF ((status EQ 0UL) OR (N_ELEMENTS(arr) LE 0L)) THEN RETURN, 0UL
         arr_size = SIZE(arr)
         arr_dims = arr_size[0]
         str_names = STRARR(8)
         FOR i=0, arr_dims-1 DO BEGIN
            temp_name = REFORM(d_names[*,i])
            DODS_FIX_NAME, temp_name
            str_names[i] = temp_name
         ENDFOR

         ; Build IDL structure to hold the components.
         var = CREATE_STRUCT('array', REFORM(TEMPORARY(arr)), $
                  'start_stride_stop', REFORM(sss[*,0:arr_dims-1]), $
                  'dimension_names', REFORM(str_names[0:arr_dims-1]))
      END
      !dods_grid_c: BEGIN
         d_names = BYTARR(65, 8)
         grid_var_p = 0UL
         status = GET_DODS_GRID(conn, dds, var_p, grid_var_p, d_names, $
                     d0, d1, d2, d3, d4, d5, d6, d7)

         IF ((status EQ 0UL) OR (grid_var_p EQ 0UL)) THEN RETURN, 0UL

         status = DODS_VAR(conn, dds, grid_var_p, grid_var, grid_name)

         IF ((status EQ 0UL) OR (N_ELEMENTS(grid_var) LE 0L)) THEN RETURN, 0UL

         str_names = STRARR(8)
         FOR i=0, 7 DO BEGIN
            temp_name = REFORM(d_names[*,i])
            DODS_FIX_NAME, temp_name
            str_names[i] = temp_name
         ENDFOR
         d_names = TEMPORARY(str_names)

         ; Build IDL structure to hold the components.

         var = CREATE_STRUCT(grid_name, REFORM(TEMPORARY(grid_var)))

         IF (N_ELEMENTS(d0) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[0])
            IF (ind[0] GE 0L) THEN d_names[0] = d_names[0] + '_1'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[0], $
                     TEMPORARY(d0))
         ENDIF
         IF (N_ELEMENTS(d1) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[1])
            IF (ind[0] GE 0L) THEN d_names[1] = d_names[1] + '_1'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[1], $
                     TEMPORARY(d1))
         ENDIF
         IF (N_ELEMENTS(d2) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[2])
            IF (ind[0] GE 0L) THEN d_names[2] = d_names[2] + '_2'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[2], $
                     TEMPORARY(d2))
         ENDIF
         IF (N_ELEMENTS(d3) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[3])
            IF (ind[0] GE 0L) THEN d_names[3] = d_names[3] + '_3'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[3], $
                     TEMPORARY(d3))
         ENDIF
         IF (N_ELEMENTS(d4) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[4])
            IF (ind[0] GE 0L) THEN d_names[4] = d_names[4] + '_4'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[4], $
                     TEMPORARY(d4))
         ENDIF
         IF (N_ELEMENTS(d5) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[5])
            IF (ind[0] GE 0L) THEN d_names[5] = d_names[5] + '_5'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[5], $
                     TEMPORARY(d5))
         ENDIF
         IF (N_ELEMENTS(d6) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[6])
            IF (ind[0] GE 0L) THEN d_names[6] = d_names[6] + '_6'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[6], $
                     TEMPORARY(d6))
         ENDIF
         IF (N_ELEMENTS(d7) GT 0L) THEN BEGIN
            ind = WHERE(TAG_NAMES(var) EQ d_names[7])
            IF (ind[0] GE 0L) THEN d_names[7] = d_names[7] + '_7'
            var = CREATE_STRUCT(TEMPORARY(var), d_names[7], $
                     TEMPORARY(d7))
         ENDIF
      END
      !dods_structure_c: BEGIN
         var_ptrs = PTR_NEW()
         names = ''
         WHILE (pix NE 0UL) DO BEGIN
            sub_var_p = 0UL
            DODS_NEXT, var_p, pix, sub_var_p
            status = DODS_VAR(conn, dds, sub_var_p, curr_var, curr_name)
            IF ((status NE 0UL) AND (N_ELEMENTS(curr_var) GT 0L)) THEN BEGIN
               var_ptrs = [TEMPORARY(var_ptrs), PTR_NEW(TEMPORARY(curr_var))]
               DODS_FIX_NAME, curr_name
               names = [TEMPORARY(names), TEMPORARY(curr_name)]
            ENDIF ELSE pix = 0UL
         ENDWHILE

         ; Build IDL structure to hold the components.

         IF (N_ELEMENTS(var_ptrs) GT 1L) THEN BEGIN
            var_ptrs = var_ptrs[1:*]
            names = names[1:*]

            var = CREATE_STRUCT(names[0], TEMPORARY(*var_ptrs[0L]))
            FOR i=1L, N_ELEMENTS(var_ptrs)-1L DO BEGIN
               ind = WHERE(TAG_NAMES(var) EQ names[i])
               IF (ind[0] GE 0L) THEN names[i] = names[i] + '_1'
               var = CREATE_STRUCT(TEMPORARY(var), $
                        names[i], TEMPORARY(*var_ptrs[i]))
            ENDFOR
         ENDIF
         PTR_FREE, var_ptrs
      END
      !dods_sequence_c: BEGIN
         var_ptrs = PTR_NEW()
         names = ''

         n_rows = 0UL
         more_seq = 1UL
         WHILE (more_seq EQ 1UL) DO BEGIN
            more_seq = DODS_SEQ(conn, dds, var_p)
            DODS_VAR_INFO, var_p, type, var_name, pix
            WHILE (pix NE 0UL) DO BEGIN
               sub_var_p = 0UL
               DODS_NEXT, var_p, pix, sub_var_p

               status = DODS_VAR(conn, dds, sub_var_p, curr_var, curr_name)
               IF ((status NE 0UL) AND (N_ELEMENTS(curr_var) GT 0L)) THEN BEGIN
                  DODS_FIX_NAME, curr_name
                  index = (WHERE(names EQ curr_name))[0]
                  curr_size = SIZE(curr_var)
                  curr_type = curr_size[curr_size[0] + 1L]
                  IF (index LT 0L) THEN BEGIN
                     IF ((N_ELEMENTS(curr_var) EQ 1L) AND (curr_type NE 8)) THEN BEGIN
                        var_ptrs = [TEMPORARY(var_ptrs), $
                                    PTR_NEW(TEMPORARY(curr_var))] 
                     ENDIF ELSE BEGIN
                        var_ptrs = [TEMPORARY(var_ptrs), $
                                    PTR_NEW(PTR_NEW(TEMPORARY(curr_var)))] 
                     ENDELSE
                     names = [TEMPORARY(names), TEMPORARY(curr_name)]
                  ENDIF ELSE BEGIN ; Concatenate existing data.
                     IF ((N_ELEMENTS(curr_var) EQ 1L) AND (curr_type NE 8)) THEN BEGIN
                        *(var_ptrs[index]) = [TEMPORARY(*(var_ptrs[index])), $
                                    PTR_NEW(TEMPORARY(curr_var))]
                     ENDIF ELSE BEGIN
                        *(var_ptrs[index]) = [TEMPORARY(*(var_ptrs[index])), $
                                    (PTR_NEW(TEMPORARY(curr_var)))]
                     ENDELSE
                  ENDELSE
               ENDIF ELSE pix = 0UL
            ENDWHILE
            n_rows = n_rows + 1UL
         ENDWHILE

         ; Build IDL structure to hold the components.

         IF (N_ELEMENTS(var_ptrs) GT 1L) THEN BEGIN
            var_ptrs = var_ptrs[1:*]
            names = names[1:*]

            var = CREATE_STRUCT(names[0], TEMPORARY(*var_ptrs[0L]))
            FOR i=1L, N_ELEMENTS(var_ptrs)-1L DO BEGIN
               ind = WHERE(TAG_NAMES(var) EQ names[i])
               IF (ind[0] GE 0L) THEN names[i] = names[i] + '_1'
               var = CREATE_STRUCT(TEMPORARY(var), $
                        names[i], TEMPORARY(*var_ptrs[i]))
            ENDFOR
         ENDIF
         PTR_FREE, var_ptrs
      END
      ELSE:
   ENDCASE

   RETURN, 1UL
END
